#define _GNU_SOURCE
#include <err.h>
#include <stdint.h>
#include <stdio.h>
#include <errno.h>
#include <semaphore.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/sysinfo.h>
#include <pthread.h>
#include <signal.h>

#define T 3

typedef struct {
    pthread_cond_t* cond;
    pthread_mutex_t* mut;
    int n;
    int cur;
    int turn;
} barrier_t;

pthread_mutex_t glob;
int barrier_code;

barrier_t barrier;

pthread_t p0_t;
void* code_block_ptr0;
pthread_t p1_t;
void* code_block_ptr1;
pthread_t p2_t;
void* code_block_ptr2;

static pthread_t* pts[] = { &p0_t , &p1_t , &p2_t   };

static uint64_t R;

void barrier_init(int n, barrier_t* b) {
    b->cond = (pthread_cond_t*)malloc(sizeof(pthread_cond_t));
    b->mut = (pthread_mutex_t*)malloc(sizeof(pthread_mutex_t));
    pthread_cond_init(b->cond, NULL) ;
    pthread_mutex_init(b->mut, NULL) ;
    b->n = n;
    b->cur = n;
    b->turn = 0;
}

void barrier_free(barrier_t* b) {
    pthread_cond_destroy(b->cond);
    pthread_mutex_destroy(b->mut);
    free(b->mut);
    free(b->cond);
}

void barrier_wait(barrier_t* bar) {
    pthread_mutex_lock(bar->mut);
    int t = bar->turn;
    bar->cur -= 1;
    if (bar->cur) {
        do {
            pthread_cond_wait(bar->cond, bar->mut);
        }
        while (bar->turn == t); /* defend against spurious wakeups */
    } else {
        bar->cur = bar->n;
        bar->turn++;
        pthread_cond_broadcast(bar->cond);
    }
    pthread_mutex_unlock(bar->mut);
}


uint64_t BLOCK_SIZE = 1L << 32L;

int randbit(void) {
    static int rn = 0;
    if (rn == 0)
        rn = rand();

    static i = 0;
    if (i < 32) {
        int k = (rn & (1 << i)) >> i;
        i++;
        return k;
    } else {
        i = 0;
        rn = rand();
        return randbit();
    }
}

struct Result {
    int64_t* p0x0;
    int64_t* p1w0;
    int64_t* p2x0;
    int64_t* p2x2;
    };

struct ResultPair {
    struct Result* result;
    int64_t count;
    int64_t validated;
};

struct Arg {
    uint64_t* my;
    uint64_t* mz;
    
    int stride;
    int** cpus;
    struct Result results[];
};

int DONE;
void* timeout(void* a) {
    int t = 0;
    while (!DONE) {
        if (t > 5000*1000 + R*100*1000) {
            printf("{\"error\":\"kill\"}\n");
            exit(7);
            return NULL;
        }
        usleep(10 * 1000);
        t += 10 * 1000;
    }
    return NULL;
}

void strongbar(void) {
    asm volatile (
        "dsb sy\n"
    : /* no outputs */
    : /* no inputs */
    : "memory"
    );
}


void set_affinity(int p) {
#ifndef NOPTHREAD
cpu_set_t cpu;
    CPU_ZERO(&cpu);
    CPU_SET(p, &cpu);
    int r;
    if (r=pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t), &cpu)) {
        errno=r;
        warn("set_affinity");
    }
#endif
}

void delay(void) {
    if (rand())
        fflush(stdout);  /* enough to force a context switch sometimes */
}

__attribute__((aligned(0x4000))) void p0fresh169(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh169) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh169:
    asm volatile (
        "b _over_chunk_codefresh169\n"
        "fresh169: nop\n"
        
        "b fresh170\n"
        "_over_chunk_codefresh169: nop\n"
    );

}__attribute__((aligned(0x4000))) void p0fresh170(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh170) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh170:
    asm volatile (
        "b _over_chunk_codefresh170\n"
        "fresh170: nop\n"
        "l1: ldr x0, [x1]\n"
        "b fresh171\n"
        "_over_chunk_codefresh170: nop\n"
    );

}__attribute__((aligned(0x4000))) void p0fresh171(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh171) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh171:
    asm volatile (
        "b _over_chunk_codefresh171\n"
        "fresh171: nop\n"
        "    l2: str w2, [x3]\n""    eor x4, x0, x0\n""    add x4, x4, x3\n""    dc cvau, x4\n""    dsb ish\n""    ic ivau, x3\n""    dsb ish\n"
        "b fresh172\n"
        "_over_chunk_codefresh171: nop\n"
    );

}__attribute__((aligned(0x4000))) void p0fresh172(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh172) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh172:
    asm volatile (
        "b _over_chunk_codefresh172\n"
        "fresh172: nop\n"
        "    l3: str x5, [x6]\n"
        "b fresh173\n"
        "_over_chunk_codefresh172: nop\n"
    );

}


void* p0(void* a) {
    strongbar();
    struct Arg* arg = (struct Arg* )a;

    p0fresh169();
    p0fresh170();
    p0fresh171();
    p0fresh172();
    

    asm ( "adr %[out], __P0_start\n" : [out] "=&r" (code_block_ptr0) : : "memory");
    //asm volatile("adr x0, \nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+0) : "x0", "w1", "memory");
    
    pthread_mutex_lock(&glob);
    barrier_code++;
    pthread_mutex_unlock(&glob);
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)code_block_ptr0 & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }

    uint32_t ldata[3];
    
    asm volatile("adr x0, l1\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+0) : "x0", "w1", "memory");
    asm volatile("adr x0, l2\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+1) : "x0", "w1", "memory");
    asm volatile("adr x0, l3\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+2) : "x0", "w1", "memory");
    

    for (int k = 0; k < R; k++) {
        
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, l1\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[0])
                errx(1, "incorrect data at label l1 in p0\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, l1\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, l2\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[1])
                errx(1, "incorrect data at label l2 in p0\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, l2\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, l3\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[2])
                errx(1, "incorrect data at label l3 in p0\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, l3\n" : : :);
            
        }
        
    }
    

    uint64_t p0x3 = 0;
    uint32_t p0w2 = 0;
    uint64_t p0x1 = 0;
    uint64_t p0x0 = 0;
    uint64_t p0x6 = 0;
    uint64_t p0x5 = 0;
    

    int c=0;
    int stride = arg->stride;
    for (int s = stride; s > 0; s--) {
        for (int k = R-s; k >= 0; k -= stride) {
    
            c++;

            p0x0 = 0;
            p0x5 = 1;
            asm ("adr %[rout], l8\n" : [rout] "=&r" (p0x3) : : );
            p0w2 = 335544321;
            p0x6 = (int64_t )&arg->my[k];
            p0x1 = (int64_t )&arg->mz[k];
            
            /* Reload data at each label */
            asm volatile (
                "adr x0, l1\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+0)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, l2\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+1)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, l3\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+2)
                : "x0", "w1", "memory"
            );
            
            

            /* Prefetch and check */
            if (rand() % 50) {
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, l1\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[0]) {
                        errx(1, "loop check: incorrect data at label l1 in p0");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, l2\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[1]) {
                        errx(1, "loop check: incorrect data at label l2 in p0");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, l3\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[2]) {
                        errx(1, "loop check: incorrect data at label l3 in p0");
                    }
                }
                
            }
            
            if (k % 10)
                set_affinity(arg->cpus[k][0]);
            barrier_wait(&barrier); // go
            delay();
        code0:
            asm volatile (
            "mov x0, %[p0x0]\n"
            
            "mov x5, %[p0x5]\n"
            
            "mov x3, %[p0x3]\n"
            
            "mov w2, %w[p0w2]\n"
            
            "mov x6, %[p0x6]\n"
            
            "mov x1, %[p0x1]\n"
            
            
            "__P0_start: nop\n"
            "b fresh169\n"
            "fresh173: nop\n"
            
            "__P0_end: nop\n"
            "mov %[p0x0], x0\n"
            
            :  [p0x0] "=&r" (p0x0)  
            :  "[p0x0]" (p0x0) ,  [p0x5] "r" (p0x5) ,  [p0x3] "r" (p0x3) ,  [p0w2] "r" (p0w2) ,  [p0x6] "r" (p0x6) ,  [p0x1] "r" (p0x1)  
            : "cc", "memory",  "x3" ,  "x4" ,  "w2" ,  "x1" ,  "x0" ,  "x6" ,  "x5"  
            );
            
            *arg->results[k].p0x0=p0x0;
            
            
            
            
            
            barrier_wait(&barrier); // go
        
            if (k % (R / 10) == 0) {
                printf(".\n"); fflush(stdout);  /* just to show it working */
            }
        
        }
    }
    if (c != R)
        err(1, "p0 not enough iterations.");
    /* reload data at each label to leave in consistent state */
    asm volatile (
        "adr x0, l1\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+0)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, l2\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+1)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, l3\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+2)
        : "x0", "w1", "memory"
    );
    
    
    strongbar();
    return NULL;
}
__attribute__((aligned(0x4000))) void p1fresh174(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh174) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh174:
    asm volatile (
        "b _over_chunk_codefresh174\n"
        "fresh174: nop\n"
        
        "b fresh175\n"
        "_over_chunk_codefresh174: nop\n"
    );

}__attribute__((aligned(0x4000))) void p1fresh175(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh175) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh175:
    asm volatile (
        "b _over_chunk_codefresh175\n"
        "fresh175: nop\n"
        "l4: ldr w0, [x2]\n"
        "b fresh176\n"
        "_over_chunk_codefresh175: nop\n"
    );

}__attribute__((aligned(0x4000))) void p1fresh176(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh176) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh176:
    asm volatile (
        "b _over_chunk_codefresh176\n"
        "fresh176: nop\n"
        "    l5: str w1, [x2]\n""    dmb st\n"
        "b fresh177\n"
        "_over_chunk_codefresh176: nop\n"
    );

}__attribute__((aligned(0x4000))) void p1fresh177(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh177) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh177:
    asm volatile (
        "b _over_chunk_codefresh177\n"
        "fresh177: nop\n"
        "    l6: str x3, [x4]\n"
        "b fresh178\n"
        "_over_chunk_codefresh177: nop\n"
    );

}


void* p1(void* a) {
    strongbar();
    struct Arg* arg = (struct Arg* )a;

    p1fresh174();
    p1fresh175();
    p1fresh176();
    p1fresh177();
    

    asm ( "adr %[out], __P1_start\n" : [out] "=&r" (code_block_ptr1) : : "memory");
    //asm volatile("adr x0, \nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+1) : "x0", "w1", "memory");
    
    pthread_mutex_lock(&glob);
    barrier_code++;
    pthread_mutex_unlock(&glob);
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)code_block_ptr1 & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }

    uint32_t ldata[3];
    
    asm volatile("adr x0, l4\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+0) : "x0", "w1", "memory");
    asm volatile("adr x0, l5\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+1) : "x0", "w1", "memory");
    asm volatile("adr x0, l6\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+2) : "x0", "w1", "memory");
    

    for (int k = 0; k < R; k++) {
        
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, l4\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[0])
                errx(1, "incorrect data at label l4 in p1\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, l4\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, l5\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[1])
                errx(1, "incorrect data at label l5 in p1\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, l5\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, l6\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[2])
                errx(1, "incorrect data at label l6 in p1\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, l6\n" : : :);
            
        }
        
    }
    

    uint64_t p1x2 = 0;
    uint64_t p1x3 = 0;
    uint32_t p1w0 = 0;
    uint32_t p1w1 = 0;
    uint64_t p1x4 = 0;
    

    int c=0;
    int stride = arg->stride;
    for (int s = stride; s > 0; s--) {
        for (int k = R-s; k >= 0; k -= stride) {
    
            c++;

            p1w0 = 0;
            p1w1 = 335544323;
            asm ("adr %[rout], l8\n" : [rout] "=&r" (p1x2) : : );
            p1x4 = (int64_t )&arg->mz[k];
            p1x3 = 1;
            
            /* Reload data at each label */
            asm volatile (
                "adr x0, l4\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+0)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, l5\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+1)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, l6\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+2)
                : "x0", "w1", "memory"
            );
            
            

            /* Prefetch and check */
            if (rand() % 50) {
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, l4\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[0]) {
                        errx(1, "loop check: incorrect data at label l4 in p1");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, l5\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[1]) {
                        errx(1, "loop check: incorrect data at label l5 in p1");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, l6\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[2]) {
                        errx(1, "loop check: incorrect data at label l6 in p1");
                    }
                }
                
            }
            
            if (k % 10)
                set_affinity(arg->cpus[k][1]);
            barrier_wait(&barrier); // go
            delay();
        code1:
            asm volatile (
            "mov w0, %w[p1w0]\n"
            
            "mov w1, %w[p1w1]\n"
            
            "mov x2, %[p1x2]\n"
            
            "mov x4, %[p1x4]\n"
            
            "mov x3, %[p1x3]\n"
            
            
            "__P1_start: nop\n"
            "b fresh174\n"
            "fresh178: nop\n"
            
            "__P1_end: nop\n"
            "mov %w[p1w0], w0\n"
            
            :  [p1w0] "=&r" (p1w0)  
            :  "[p1w0]" (p1w0) ,  [p1w1] "r" (p1w1) ,  [p1x2] "r" (p1x2) ,  [p1x4] "r" (p1x4) ,  [p1x3] "r" (p1x3)  
            : "cc", "memory",  "x2" ,  "x3" ,  "w0" ,  "w1" ,  "x4"  
            );
            
            p1w0 = p1w0 & 0x00000000ffffffff;
            
            *arg->results[k].p1w0=p1w0;
            
            
            
            
            barrier_wait(&barrier); // go
        
        }
    }
    if (c != R)
        err(1, "p1 not enough iterations.");
    /* reload data at each label to leave in consistent state */
    asm volatile (
        "adr x0, l4\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+0)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, l5\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+1)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, l6\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+2)
        : "x0", "w1", "memory"
    );
    
    
    strongbar();
    return NULL;
}
__attribute__((aligned(0x4000))) void p2fresh179(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh179) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh179:
    asm volatile (
        "b _over_chunk_codefresh179\n"
        "fresh179: nop\n"
        
        "b fresh180\n"
        "_over_chunk_codefresh179: nop\n"
    );

}__attribute__((aligned(0x4000))) void p2fresh180(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh180) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh180:
    asm volatile (
        "b _over_chunk_codefresh180\n"
        "fresh180: nop\n"
        "l7: ldr x0, [x1]\n""    cbz x0, lc00\n"
        "b fresh181\n"
        "_over_chunk_codefresh180: nop\n"
    );

}__attribute__((aligned(0x4000))) void p2fresh181(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh181) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh181:
    asm volatile (
        "b _over_chunk_codefresh181\n"
        "fresh181: nop\n"
        "    lc00: isb\n"
        "b fresh182\n"
        "_over_chunk_codefresh181: nop\n"
    );

}__attribute__((aligned(0x4000))) void p2fresh182(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh182) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh182:
    asm volatile (
        "b _over_chunk_codefresh182\n"
        "fresh182: nop\n"
        "    l8: b fail_l8\n""    mov x2, #2\n""    b end\n""    mov x2, #3\n""    b end\n"
        "b fresh183\n"
        "_over_chunk_codefresh182: nop\n"
    );

}__attribute__((aligned(0x4000))) void p2fresh183(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh183) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh183:
    asm volatile (
        "b _over_chunk_codefresh183\n"
        "fresh183: nop\n"
        "    fail_l8: mov x2, #1\n"
        "b fresh184\n"
        "_over_chunk_codefresh183: nop\n"
    );

}__attribute__((aligned(0x4000))) void p2fresh184(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh184) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh184:
    asm volatile (
        "b _over_chunk_codefresh184\n"
        "fresh184: nop\n"
        "    end: nop\n"
        "b fresh185\n"
        "_over_chunk_codefresh184: nop\n"
    );

}


void* p2(void* a) {
    strongbar();
    struct Arg* arg = (struct Arg* )a;

    p2fresh179();
    p2fresh180();
    p2fresh181();
    p2fresh182();
    p2fresh183();
    p2fresh184();
    

    asm ( "adr %[out], __P2_start\n" : [out] "=&r" (code_block_ptr2) : : "memory");
    //asm volatile("adr x0, \nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+2) : "x0", "w1", "memory");
    
    pthread_mutex_lock(&glob);
    barrier_code++;
    pthread_mutex_unlock(&glob);
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)code_block_ptr2 & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }

    uint32_t ldata[5];
    
    asm volatile("adr x0, l7\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+0) : "x0", "w1", "memory");
    asm volatile("adr x0, lc00\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+1) : "x0", "w1", "memory");
    asm volatile("adr x0, l8\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+2) : "x0", "w1", "memory");
    asm volatile("adr x0, fail_l8\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+3) : "x0", "w1", "memory");
    asm volatile("adr x0, end\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+4) : "x0", "w1", "memory");
    

    for (int k = 0; k < R; k++) {
        
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, l7\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[0])
                errx(1, "incorrect data at label l7 in p2\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, l7\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, lc00\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[1])
                errx(1, "incorrect data at label lc00 in p2\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, lc00\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, l8\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[2])
                errx(1, "incorrect data at label l8 in p2\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, l8\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, fail_l8\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[3])
                errx(1, "incorrect data at label fail_l8 in p2\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, fail_l8\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, end\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[4])
                errx(1, "incorrect data at label end in p2\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, end\n" : : :);
            
        }
        
    }
    

    uint64_t p2x1 = 0;
    uint64_t p2x0 = 0;
    uint64_t p2x2 = 0;
    

    int c=0;
    int stride = arg->stride;
    for (int s = stride; s > 0; s--) {
        for (int k = R-s; k >= 0; k -= stride) {
    
            c++;

            p2x1 = (int64_t )&arg->my[k];
            p2x2 = 0;
            p2x0 = 0;
            
            /* Reload data at each label */
            asm volatile (
                "adr x0, l7\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+0)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, lc00\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+1)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, l8\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+2)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, fail_l8\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+3)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, end\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+4)
                : "x0", "w1", "memory"
            );
            
            

            /* Prefetch and check */
            if (rand() % 50) {
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, l7\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[0]) {
                        errx(1, "loop check: incorrect data at label l7 in p2");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, lc00\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[1]) {
                        errx(1, "loop check: incorrect data at label lc00 in p2");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, l8\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[2]) {
                        errx(1, "loop check: incorrect data at label l8 in p2");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, fail_l8\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[3]) {
                        errx(1, "loop check: incorrect data at label fail_l8 in p2");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, end\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[4]) {
                        errx(1, "loop check: incorrect data at label end in p2");
                    }
                }
                
            }
            
            if (k % 10)
                set_affinity(arg->cpus[k][2]);
            barrier_wait(&barrier); // go
            delay();
        code2:
            asm volatile (
            "mov x1, %[p2x1]\n"
            
            "mov x2, %[p2x2]\n"
            
            "mov x0, %[p2x0]\n"
            
            
            "__P2_start: nop\n"
            "b fresh179\n"
            "fresh185: nop\n"
            
            "__P2_end: nop\n"
            "mov %[p2x0], x0\n"
            "mov %[p2x2], x2\n"
            
            :  [p2x0] "=&r" (p2x0) ,  [p2x2] "=&r" (p2x2)  
            :  [p2x1] "r" (p2x1) ,  "[p2x2]" (p2x2) ,  "[p2x0]" (p2x0)  
            : "cc", "memory",  "x1" ,  "x0" ,  "x2"  
            );
            
            
            
            *arg->results[k].p2x0=p2x0;
            
            
            *arg->results[k].p2x2=p2x2;
            
            
            barrier_wait(&barrier); // go
        
        }
    }
    if (c != R)
        err(1, "p2 not enough iterations.");
    /* reload data at each label to leave in consistent state */
    asm volatile (
        "adr x0, l7\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+0)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, lc00\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+1)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, l8\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+2)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, fail_l8\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+3)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, end\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+4)
        : "x0", "w1", "memory"
    );
    
    
    strongbar();
    return NULL;
}


void shuffle(int size, void** arr) {
    int k;
    void* tmp;
    for (int i = 0; i < size; i++) {
        k = rand() % size;
        tmp = arr[k];
        arr[k] = arr[i];
        arr[i] = tmp;
    }
}

void ishuffle(int size, int* arr) {
    int k;
    int tmp;
    for (int i = 0; i < size; i++) {
        k = rand() % size;
        tmp = arr[k];
        arr[k] = arr[i];
        arr[i] = tmp;
    }
}


pthread_t hammer0;
pthread_t hammer1;
pthread_t hammer2;

uint32_t* hammer_instr_block;

struct hammer_arg {
    int* flag;
    void** proc;
};

int64_t MOD(int64_t n, int64_t k) {
    if (n < 0)
        return (k - ((-n) % k)) % k;
    return n % k;
}

void* hammer(void* a) {
    /* to hammer the cores but also to fool branch prediction/branch target prediction somewhat
     * it's common for CPUs to have a generic branch predictor which uses only the lower n bits of the VA
     * to predict on [target/taken]. We can abuse that by placing a process with the same lower n bit address space in a different page
     * that randomly jumps about while the other threads are working...
     *
     * We do this by creating a list of jump targets, shuffling them, converting each to an instruction that jumps to that target then replacing one of them with a RET.
     * All we then need to do is BL to the first jump and we're done.
     *
     * (N.B. It's not clear that this helps that much.)
     */
    struct I {
        int pos;
        struct I* next;
    };

    struct hammer_arg* arg = (struct hammer_arg* )a;
    while (1) {
        pthread_mutex_lock(&glob);
        if (barrier_code >= T) {
            pthread_mutex_unlock(&glob);
            break;
        }
        pthread_mutex_unlock(&glob);
    }

    // Now fill an 8k block with our branches
    // at instr_block+(arg->proc & 0xffffffff)
    int n = 1 << 16;
    int i;
    struct I* locs[n];
    uint64_t code_ptr = (uint64_t) *(arg->proc);
    uint64_t hammer_ptr = (uint64_t) hammer_instr_block;
    uint64_t hammer_ptr_lower = hammer_ptr & 0xffffffffL;
    int64_t code_ptr_lower = (code_ptr & 0xffffffffL);
    int64_t k_ptr = MOD(code_ptr_lower - hammer_ptr_lower, BLOCK_SIZE);
    int64_t k = k_ptr / sizeof(int32_t);

    for (i = 0; i < n; i++) {
        struct I* s = (struct I*)malloc(sizeof(struct I));
        s->pos = 0;
        s->next = NULL;
        locs[i] = s;
    }

    int m;
    while (*arg->flag) {
        m = 1 + (rand() % (n - 1));

        for (i = 1; i < m; i++) {
            locs[i - 1]->next = locs[i];
        }

        locs[m - 1]->next = locs[0];

        shuffle(m, (void**)locs);

        for (i = 0; i < m; i++) {
            locs[i]->pos = i;
        }

        for (i = 0; i < m; i++) {
            int32_t offset = locs[i]->next->pos - i;
            uint32_t signed_offset = (uint32_t)offset & 0x3ffffff;
            
            hammer_instr_block[k+i] = (5 << 26) | signed_offset;
            
        }

        
        hammer_instr_block[k+m - 1] = 3596551104;  // RET
        

        delay();

        for (int i = 0; i < m; i++) {
        
            asm volatile (
                    "DC CVAU, %[jumper]\n"
                    "DSB ISH\n"
                    "IC IVAU, %[jumper]\n"
                    "DSB ISH\n"
            :
            : [jumper] "r" (hammer_instr_block+k+i)
            : "memory"
            );
        
        }

        asm volatile (
        
            "ISB\n"
            "BLR %[jumper]\n"
        
        :
        : [jumper] "r" (hammer_instr_block+k)
        :);

        delay();
    }


    return NULL;
}


int validate_results(struct Arg* arg) {
    int witnesses = 0;
    for (int n = 0; n < R; n++) {
        uint64_t p0x0 = *arg->results[n].p0x0;
        uint32_t p1w0 = *arg->results[n].p1w0;
        uint64_t p2x0 = *arg->results[n].p2x0;
        uint64_t p2x2 = *arg->results[n].p2x2;
        uint32_t lbl_l1;
        asm volatile("adr x0, l1\nldr %[data], [x0]\n" : [data] "=r" (lbl_l1) : : "x0", "memory");
        
        uint32_t lbl_l2;
        asm volatile("adr x0, l2\nldr %[data], [x0]\n" : [data] "=r" (lbl_l2) : : "x0", "memory");
        
        uint32_t lbl_l3;
        asm volatile("adr x0, l3\nldr %[data], [x0]\n" : [data] "=r" (lbl_l3) : : "x0", "memory");
        
        uint32_t lbl_l4;
        asm volatile("adr x0, l4\nldr %[data], [x0]\n" : [data] "=r" (lbl_l4) : : "x0", "memory");
        
        uint32_t lbl_l5;
        asm volatile("adr x0, l5\nldr %[data], [x0]\n" : [data] "=r" (lbl_l5) : : "x0", "memory");
        
        uint32_t lbl_l6;
        asm volatile("adr x0, l6\nldr %[data], [x0]\n" : [data] "=r" (lbl_l6) : : "x0", "memory");
        
        uint32_t lbl_l7;
        asm volatile("adr x0, l7\nldr %[data], [x0]\n" : [data] "=r" (lbl_l7) : : "x0", "memory");
        
        uint32_t lbl_lc00;
        asm volatile("adr x0, lc00\nldr %[data], [x0]\n" : [data] "=r" (lbl_lc00) : : "x0", "memory");
        
        uint32_t lbl_l8;
        asm volatile("adr x0, l8\nldr %[data], [x0]\n" : [data] "=r" (lbl_l8) : : "x0", "memory");
        
        uint32_t lbl_fail_l8;
        asm volatile("adr x0, fail_l8\nldr %[data], [x0]\n" : [data] "=r" (lbl_fail_l8) : : "x0", "memory");
        
        uint32_t lbl_end;
        asm volatile("adr x0, end\nldr %[data], [x0]\n" : [data] "=r" (lbl_end) : : "x0", "memory");
        
        
        if (p0x0 == 1) {
            
        if (p1w0 == 335544321) {
            
        if (p2x0 == 1) {
            
        if (p2x2 == 2) {
            witnesses++;
        }
        
        }
        
        }
        
        }
        
    }
    return witnesses;
}

struct Arg* mkArg(void) {
    struct Arg* arg = (struct Arg*)malloc(sizeof(struct Arg) + sizeof(struct Result[R]));
    int** allcpus = (int**)malloc(sizeof(int*)*R);
    arg->cpus = allcpus;
int nprocs = get_nprocs();

    for (int k = 0; k < R; k++) {
        int* cpus = (int*)malloc(sizeof(int)*T);
        int start = rand() % nprocs;
        for (int i = 0; i < T; i++) {
            int top = rand() % 2;
            if (top == 0) {
                int b = start - (start % 8);
                start = b + rand() % 8;
            }
            else {
                int b = start - (start % 8);
                start = b + 8*(rand() % nprocs) + rand() % 8;
            }
            start = (nprocs + start) % nprocs;
            cpus[i] = start;
        }
        ishuffle(T, cpus);
        allcpus[k] = cpus;

        arg->results[k].p0x0 = (int64_t*) malloc(sizeof(int64_t));
        arg->results[k].p1w0 = (int64_t*) malloc(sizeof(int64_t));
        arg->results[k].p2x0 = (int64_t*) malloc(sizeof(int64_t));
        arg->results[k].p2x2 = (int64_t*) malloc(sizeof(int64_t));
        }
    shuffle(R, (void**)allcpus);

    int stride = rand() % R;
    arg->stride = stride;

    arg->my = (uint64_t*)malloc(sizeof(uint64_t)*R);
    for (int i = 0; i < R; i++) {
        arg->my[i] = 0;
    }
    arg->mz = (uint64_t*)malloc(sizeof(uint64_t)*R);
    for (int i = 0; i < R; i++) {
        arg->mz[i] = 0;
    }
    

    return arg;
}

int main(int argc,char **argv) {
    // allocate a 512MiB space that we will fill with our branches at the designated locations
    hammer_instr_block = (uint32_t* )mmap(NULL, BLOCK_SIZE, PROT_READ|PROT_EXEC|PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0);
    
    srand(time(NULL));

    if (argc > 1)
        R = atoi(argv[1]);
    else
        R = 1;

    int K = 4;
    struct Arg* arg = mkArg();

    barrier_init(T, &barrier);

    int flag = 1;
    int r;
    cpu_set_t hammer_cpus0;
    struct hammer_arg hammer_arg0;
    hammer_arg0.proc = &code_block_ptr0;
    hammer_arg0.flag = &flag;
    pthread_create(&hammer0, NULL, *hammer, (void* )&hammer_arg0);

    CPU_ZERO(&hammer_cpus0);
    CPU_SET(arg->cpus[0][0], &hammer_cpus0);
    if (r=pthread_setaffinity_np(hammer0, sizeof(cpu_set_t), &hammer_cpus0)) {
        errno=r;
        warn("setaffinity_hammer0");
    }
    cpu_set_t hammer_cpus1;
    struct hammer_arg hammer_arg1;
    hammer_arg1.proc = &code_block_ptr1;
    hammer_arg1.flag = &flag;
    pthread_create(&hammer1, NULL, *hammer, (void* )&hammer_arg1);

    CPU_ZERO(&hammer_cpus1);
    CPU_SET(arg->cpus[0][1], &hammer_cpus1);
    if (r=pthread_setaffinity_np(hammer1, sizeof(cpu_set_t), &hammer_cpus1)) {
        errno=r;
        warn("setaffinity_hammer1");
    }
    cpu_set_t hammer_cpus2;
    struct hammer_arg hammer_arg2;
    hammer_arg2.proc = &code_block_ptr2;
    hammer_arg2.flag = &flag;
    pthread_create(&hammer2, NULL, *hammer, (void* )&hammer_arg2);

    CPU_ZERO(&hammer_cpus2);
    CPU_SET(arg->cpus[0][2], &hammer_cpus2);
    if (r=pthread_setaffinity_np(hammer2, sizeof(cpu_set_t), &hammer_cpus2)) {
        errno=r;
        warn("setaffinity_hammer2");
    }
    
    

    DONE = 0;
    pthread_t timeout_t;
    pthread_create(&timeout_t, NULL, *timeout, NULL);

    void (*fps[T]);
    
    fps[0]=p0;
    fps[1]=p1;
    fps[2]=p2;
    

    shuffle(T, (void**)fps);
    shuffle(T, (void**)pts);
    for (int i = 0; i < T; i++) {
        pthread_create(pts[i], NULL, fps[i], (void* )arg);
    }

    shuffle(T, (void**)pts);
    for (int i = 0; i < T; i++) {
        pthread_join(*(pts[i]), NULL);
    }

    flag = 0;
    pthread_join(hammer0, NULL);
    pthread_join(hammer1, NULL);
    pthread_join(hammer2, NULL);
    
    
    DONE = 1;
    pthread_join(timeout_t, NULL);
    /* collect results */
    struct ResultPair count[R];
    int jfill = 0;
    int found = 0;
    for (int k = 0; k < R; k++) {
        for (int j = 0; j < jfill; j++) {
            if (*count[j].result->p0x0 != *arg->results[k].p0x0)
                 continue;
            if (*count[j].result->p1w0 != *arg->results[k].p1w0)
                 continue;
            if (*count[j].result->p2x0 != *arg->results[k].p2x0)
                 continue;
            if (*count[j].result->p2x2 != *arg->results[k].p2x2)
                 continue;
            count[j].count++;
            found = 1;
        }

        if (found == 0) {
            count[jfill].result = &(arg->results[k]);
            count[jfill].count=1;
            jfill++;
        }
        found = 0;
    }

    int witnesses = validate_results(arg);
    for (int k = 0; k < jfill; k++) {
        char* prefix = "";
        printf("{ %s \"p0x0\":%lu,\"p1w0\":%u,\"p2x0\":%lu,\"p2x2\":%lu} : %lu\n", prefix,  *count[k].result->p0x0 ,  *count[k].result->p1w0 ,  *count[k].result->p2x0 ,  *count[k].result->p2x2  , count[k].count);
    }
    printf("WITNESS: %d\n", witnesses);
    barrier_free(&barrier);
    munmap((void*)hammer_instr_block, BLOCK_SIZE);
    
    free(arg->my);
    free(arg->mz);
    
    for (int k = 0; k < R; k++)
    {
        free(arg->results[k].p0x0);
        free(arg->results[k].p1w0);
        free(arg->results[k].p2x0);
        free(arg->results[k].p2x2);
        }
    free(arg);
    return 0;
}