#define _GNU_SOURCE
#include <err.h>
#include <stdint.h>
#include <stdio.h>
#include <errno.h>
#include <semaphore.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/sysinfo.h>
#include <pthread.h>
#include <signal.h>

#define T 3

typedef struct {
    pthread_cond_t* cond;
    pthread_mutex_t* mut;
    int n;
    int cur;
    int turn;
} barrier_t;

pthread_mutex_t glob;
int barrier_code;

barrier_t barrier;

pthread_t p0_t;
void* code_block_ptr0;
pthread_t p1_t;
void* code_block_ptr1;
pthread_t p2_t;
void* code_block_ptr2;

static pthread_t* pts[] = { &p0_t , &p1_t , &p2_t   };

static uint64_t R;

void barrier_init(int n, barrier_t* b) {
    b->cond = (pthread_cond_t*)malloc(sizeof(pthread_cond_t));
    b->mut = (pthread_mutex_t*)malloc(sizeof(pthread_mutex_t));
    pthread_cond_init(b->cond, NULL) ;
    pthread_mutex_init(b->mut, NULL) ;
    b->n = n;
    b->cur = n;
    b->turn = 0;
}

void barrier_free(barrier_t* b) {
    pthread_cond_destroy(b->cond);
    pthread_mutex_destroy(b->mut);
    free(b->mut);
    free(b->cond);
}

void barrier_wait(barrier_t* bar) {
    pthread_mutex_lock(bar->mut);
    int t = bar->turn;
    bar->cur -= 1;
    if (bar->cur) {
        do {
            pthread_cond_wait(bar->cond, bar->mut);
        }
        while (bar->turn == t); /* defend against spurious wakeups */
    } else {
        bar->cur = bar->n;
        bar->turn++;
        pthread_cond_broadcast(bar->cond);
    }
    pthread_mutex_unlock(bar->mut);
}


uint64_t BLOCK_SIZE = 1L << 32L;

int randbit(void) {
    static int rn = 0;
    if (rn == 0)
        rn = rand();

    static i = 0;
    if (i < 32) {
        int k = (rn & (1 << i)) >> i;
        i++;
        return k;
    } else {
        i = 0;
        rn = rand();
        return randbit();
    }
}

struct Result {
    int64_t* p2x3;
    int64_t* p2x2;
    int64_t* p1x2;
    };

struct ResultPair {
    struct Result* result;
    int64_t count;
    int64_t validated;
};

struct Arg {
    uint64_t* mz;
    
    int stride;
    int** cpus;
    struct Result results[];
};

int DONE;
void* timeout(void* a) {
    int t = 0;
    while (!DONE) {
        if (t > 5000*1000 + R*100*1000) {
            printf("{\"error\":\"kill\"}\n");
            exit(7);
            return NULL;
        }
        usleep(10 * 1000);
        t += 10 * 1000;
    }
    return NULL;
}

void strongbar(void) {
    asm volatile (
        "dsb sy\n"
    : /* no outputs */
    : /* no inputs */
    : "memory"
    );
}


void set_affinity(int p) {
#ifndef NOPTHREAD
cpu_set_t cpu;
    CPU_ZERO(&cpu);
    CPU_SET(p, &cpu);
    int r;
    if (r=pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t), &cpu)) {
        errno=r;
        warn("set_affinity");
    }
#endif
}

void delay(void) {
    if (rand())
        fflush(stdout);  /* enough to force a context switch sometimes */
}

__attribute__((aligned(0x4000))) void p0fresh104(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh104) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh104:
    asm volatile (
        "b _over_chunk_codefresh104\n"
        "fresh104: nop\n"
        "str w0, [x1]\n"
        "b fresh105\n"
        "_over_chunk_codefresh104: nop\n"
    );

}


void* p0(void* a) {
    strongbar();
    struct Arg* arg = (struct Arg* )a;

    p0fresh104();
    

    asm ( "adr %[out], __P0_start\n" : [out] "=&r" (code_block_ptr0) : : "memory");
    //asm volatile("adr x0, \nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+0) : "x0", "w1", "memory");
    
    pthread_mutex_lock(&glob);
    barrier_code++;
    pthread_mutex_unlock(&glob);
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)code_block_ptr0 & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }

    
    

    for (int k = 0; k < R; k++) {
        
        
    }
    

    uint64_t p0x1 = 0;
    uint32_t p0w0 = 0;
    

    int c=0;
    int stride = arg->stride;
    for (int s = stride; s > 0; s--) {
        for (int k = R-s; k >= 0; k -= stride) {
    
            c++;

            asm ("adr %[rout], a\n" : [rout] "=&r" (p0x1) : : );
            p0w0 = 3573751839;
            
            /* Reload data at each label */
            

            /* Prefetch and check */
            if (rand() % 50) {
                
            }
            
            if (k % 10)
                set_affinity(arg->cpus[k][0]);
            barrier_wait(&barrier); // go
            delay();
        code0:
            asm volatile (
            "mov x1, %[p0x1]\n"
            
            "mov w0, %w[p0w0]\n"
            
            
            "__P0_start: nop\n"
            "b fresh104\n"
            "fresh105: nop\n"
            
            "__P0_end: nop\n"
            
            : 
            :  [p0x1] "r" (p0x1) ,  [p0w0] "r" (p0w0)  
            : "cc", "memory",  "x1" ,  "w0"  
            );
            
            
            
            
            barrier_wait(&barrier); // go
        
            if (k % (R / 10) == 0) {
                printf(".\n"); fflush(stdout);  /* just to show it working */
            }
        
        }
    }
    if (c != R)
        err(1, "p0 not enough iterations.");
    /* reload data at each label to leave in consistent state */
    
    strongbar();
    return NULL;
}
__attribute__((aligned(0x4000))) void p1fresh106(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh106) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh106:
    asm volatile (
        "b _over_chunk_codefresh106\n"
        "fresh106: nop\n"
        "bl a\n""    b end\n"
        "b fresh107\n"
        "_over_chunk_codefresh106: nop\n"
    );

}__attribute__((aligned(0x4000))) void p1fresh107(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh107) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh107:
    asm volatile (
        "b _over_chunk_codefresh107\n"
        "fresh107: nop\n"
        "    a: b over\n""    mov x2, #1\n""    str w0, [x1]\n"
        "b fresh108\n"
        "_over_chunk_codefresh107: nop\n"
    );

}__attribute__((aligned(0x4000))) void p1fresh108(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh108) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh108:
    asm volatile (
        "b _over_chunk_codefresh108\n"
        "fresh108: nop\n"
        "    over: ret\n"
        "b fresh109\n"
        "_over_chunk_codefresh108: nop\n"
    );

}__attribute__((aligned(0x4000))) void p1fresh109(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh109) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh109:
    asm volatile (
        "b _over_chunk_codefresh109\n"
        "fresh109: nop\n"
        "    end: nop\n"
        "b fresh110\n"
        "_over_chunk_codefresh109: nop\n"
    );

}


void* p1(void* a) {
    strongbar();
    struct Arg* arg = (struct Arg* )a;

    p1fresh106();
    p1fresh107();
    p1fresh108();
    p1fresh109();
    

    asm ( "adr %[out], __P1_start\n" : [out] "=&r" (code_block_ptr1) : : "memory");
    //asm volatile("adr x0, \nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+1) : "x0", "w1", "memory");
    
    pthread_mutex_lock(&glob);
    barrier_code++;
    pthread_mutex_unlock(&glob);
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)code_block_ptr1 & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }

    uint32_t ldata[3];
    
    asm volatile("adr x0, a\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+0) : "x0", "w1", "memory");
    asm volatile("adr x0, over\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+1) : "x0", "w1", "memory");
    asm volatile("adr x0, end\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+2) : "x0", "w1", "memory");
    

    for (int k = 0; k < R; k++) {
        
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, a\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[0])
                errx(1, "incorrect data at label a in p1\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, a\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, over\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[1])
                errx(1, "incorrect data at label over in p1\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, over\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, end\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[2])
                errx(1, "incorrect data at label end in p1\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, end\n" : : :);
            
        }
        
    }
    

    uint64_t p1x2 = 0;
    uint32_t p1w0 = 0;
    uint64_t p1x1 = 0;
    

    int c=0;
    int stride = arg->stride;
    for (int s = stride; s > 0; s--) {
        for (int k = R-s; k >= 0; k -= stride) {
    
            c++;

            asm ("adr %[rout], l1\n" : [rout] "=&r" (p1x1) : : );
            p1x2 = 0;
            p1w0 = 3573751839;
            
            /* Reload data at each label */
            asm volatile (
                "adr x0, a\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+0)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, over\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+1)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, end\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+2)
                : "x0", "w1", "memory"
            );
            
            

            /* Prefetch and check */
            if (rand() % 50) {
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, a\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[0]) {
                        errx(1, "loop check: incorrect data at label a in p1");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, over\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[1]) {
                        errx(1, "loop check: incorrect data at label over in p1");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, end\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[2]) {
                        errx(1, "loop check: incorrect data at label end in p1");
                    }
                }
                
            }
            
            if (k % 10)
                set_affinity(arg->cpus[k][1]);
            barrier_wait(&barrier); // go
            delay();
        code1:
            asm volatile (
            "mov x1, %[p1x1]\n"
            
            "mov x2, %[p1x2]\n"
            
            "mov w0, %w[p1w0]\n"
            
            
            "__P1_start: nop\n"
            "b fresh106\n"
            "fresh110: nop\n"
            
            "__P1_end: nop\n"
            "mov %[p1x2], x2\n"
            
            :  [p1x2] "=&r" (p1x2)  
            :  [p1x1] "r" (p1x1) ,  "[p1x2]" (p1x2) ,  [p1w0] "r" (p1w0)  
            : "cc", "memory",  "x2" ,  "w0" ,  "x1"  
            );
            
            
            
            *arg->results[k].p1x2=p1x2;
            
            
            barrier_wait(&barrier); // go
        
        }
    }
    if (c != R)
        err(1, "p1 not enough iterations.");
    /* reload data at each label to leave in consistent state */
    asm volatile (
        "adr x0, a\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+0)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, over\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+1)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, end\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+2)
        : "x0", "w1", "memory"
    );
    
    
    strongbar();
    return NULL;
}
__attribute__((aligned(0x4000))) void p2fresh111(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh111) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh111:
    asm volatile (
        "b _over_chunk_codefresh111\n"
        "fresh111: nop\n"
        
        "b fresh112\n"
        "_over_chunk_codefresh111: nop\n"
    );

}__attribute__((aligned(0x4000))) void p2fresh112(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh112) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh112:
    asm volatile (
        "b _over_chunk_codefresh112\n"
        "fresh112: nop\n"
        "l1: b end2\n""    mov x3, #1\n"
        "b fresh113\n"
        "_over_chunk_codefresh112: nop\n"
    );

}__attribute__((aligned(0x4000))) void p2fresh113(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh113) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh113:
    asm volatile (
        "b _over_chunk_codefresh113\n"
        "fresh113: nop\n"
        "    l2: bl a\n"
        "b fresh114\n"
        "_over_chunk_codefresh113: nop\n"
    );

}__attribute__((aligned(0x4000))) void p2fresh114(void){
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)(&&_chunk_codefresh114) & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }


_chunk_codefresh114:
    asm volatile (
        "b _over_chunk_codefresh114\n"
        "fresh114: nop\n"
        "    end2: nop\n"
        "b fresh115\n"
        "_over_chunk_codefresh114: nop\n"
    );

}


void* p2(void* a) {
    strongbar();
    struct Arg* arg = (struct Arg* )a;

    p2fresh111();
    p2fresh112();
    p2fresh113();
    p2fresh114();
    

    asm ( "adr %[out], __P2_start\n" : [out] "=&r" (code_block_ptr2) : : "memory");
    //asm volatile("adr x0, \nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+2) : "x0", "w1", "memory");
    
    pthread_mutex_lock(&glob);
    barrier_code++;
    pthread_mutex_unlock(&glob);
    int64_t pagesize = getpagesize();
    void* code_block_ptr = (void*)((long)code_block_ptr2 & ~(pagesize-1));
    if(mprotect(code_block_ptr, pagesize, PROT_READ|PROT_EXEC|PROT_WRITE)) {
        err(1, "mprotect");
    }

    uint32_t ldata[3];
    
    asm volatile("adr x0, l1\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+0) : "x0", "w1", "memory");
    asm volatile("adr x0, l2\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+1) : "x0", "w1", "memory");
    asm volatile("adr x0, end2\nldr w1, [x0]\nstr w1, [%[data]]" : : [data] "r" (ldata+2) : "x0", "w1", "memory");
    

    for (int k = 0; k < R; k++) {
        
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, l1\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[0])
                errx(1, "incorrect data at label l1 in p2\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, l1\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, l2\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[1])
                errx(1, "incorrect data at label l2 in p2\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, l2\n" : : :);
            
        }
        
        if (randbit()) {
            uint32_t xi;
            asm volatile ("adr x0, end2\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "x1", "memory");
            
            if (xi != ldata[2])
                errx(1, "incorrect data at label end2 in p2\n");
        }
        if (randbit()) {
            asm volatile ("prfm plil1keep, end2\n" : : :);
            
        }
        
    }
    

    uint64_t p2x1 = 0;
    uint64_t p2x3 = 0;
    uint32_t p2w0 = 0;
    uint64_t p2x2 = 0;
    

    int c=0;
    int stride = arg->stride;
    for (int s = stride; s > 0; s--) {
        for (int k = R-s; k >= 0; k -= stride) {
    
            c++;

            p2x2 = 0;
            p2w0 = 3573751839;
            asm ("adr %[rout], l1\n" : [rout] "=&r" (p2x1) : : );
            
            /* Reload data at each label */
            asm volatile (
                "adr x0, l1\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+0)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, l2\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+1)
                : "x0", "w1", "memory"
            );
            
            asm volatile (
                "adr x0, end2\n"
                "ldr w1, [%[ldata]]\n"
                "str w1, [x0]\n"
                "dc cvau, x0\n"
                "dsb ish\n"
                "ic ivau, x0\n"
                "dsb ish\n"
                :
                : [ldata] "r" (ldata+2)
                : "x0", "w1", "memory"
            );
            
            

            /* Prefetch and check */
            if (rand() % 50) {
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, l1\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[0]) {
                        errx(1, "loop check: incorrect data at label l1 in p2");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, l2\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[1]) {
                        errx(1, "loop check: incorrect data at label l2 in p2");
                    }
                }
                
                if (rand() % 2) {
                    uint32_t xi;
                    
                    asm volatile ("adr x0, end2\nldr %w[x1], [x0]\n" : [x1] "=&r" (xi) : : "x0", "memory");
                    
                    if (xi != ldata[2]) {
                        errx(1, "loop check: incorrect data at label end2 in p2");
                    }
                }
                
            }
            
            if (k % 10)
                set_affinity(arg->cpus[k][2]);
            barrier_wait(&barrier); // go
            delay();
        code2:
            asm volatile (
            "mov x2, %[p2x2]\n"
            
            "mov w0, %w[p2w0]\n"
            
            "mov x1, %[p2x1]\n"
            
            
            "__P2_start: nop\n"
            "b fresh111\n"
            "fresh115: nop\n"
            
            "__P2_end: nop\n"
            "mov %[p2x3], x3\n"
            "mov %[p2x2], x2\n"
            
            :  [p2x3] "=&r" (p2x3) ,  [p2x2] "=&r" (p2x2)  
            :  "[p2x2]" (p2x2) ,  [p2w0] "r" (p2w0) ,  [p2x1] "r" (p2x1)  
            : "cc", "memory",  "x1" ,  "x3" ,  "w0" ,  "x2"  
            );
            
            *arg->results[k].p2x3=p2x3;
            
            
            *arg->results[k].p2x2=p2x2;
            
            
            
            barrier_wait(&barrier); // go
        
        }
    }
    if (c != R)
        err(1, "p2 not enough iterations.");
    /* reload data at each label to leave in consistent state */
    asm volatile (
        "adr x0, l1\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+0)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, l2\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+1)
        : "x0", "w1", "memory"
    );
    
    asm volatile (
        "adr x0, end2\n"
        "ldr w1, [%[ldata]]\n"
        "str w1, [x0]\n"
        "dc cvau, x0\n"
        "dsb ish\n"
        "ic ivau, x0\n"
        "dsb ish\n"
        :
        : [ldata] "r" (ldata+2)
        : "x0", "w1", "memory"
    );
    
    
    strongbar();
    return NULL;
}


void shuffle(int size, void** arr) {
    int k;
    void* tmp;
    for (int i = 0; i < size; i++) {
        k = rand() % size;
        tmp = arr[k];
        arr[k] = arr[i];
        arr[i] = tmp;
    }
}

void ishuffle(int size, int* arr) {
    int k;
    int tmp;
    for (int i = 0; i < size; i++) {
        k = rand() % size;
        tmp = arr[k];
        arr[k] = arr[i];
        arr[i] = tmp;
    }
}


pthread_t hammer0;
pthread_t hammer1;
pthread_t hammer2;

uint32_t* hammer_instr_block;

struct hammer_arg {
    int* flag;
    void** proc;
};

int64_t MOD(int64_t n, int64_t k) {
    if (n < 0)
        return (k - ((-n) % k)) % k;
    return n % k;
}

void* hammer(void* a) {
    /* to hammer the cores but also to fool branch prediction/branch target prediction somewhat
     * it's common for CPUs to have a generic branch predictor which uses only the lower n bits of the VA
     * to predict on [target/taken]. We can abuse that by placing a process with the same lower n bit address space in a different page
     * that randomly jumps about while the other threads are working...
     *
     * We do this by creating a list of jump targets, shuffling them, converting each to an instruction that jumps to that target then replacing one of them with a RET.
     * All we then need to do is BL to the first jump and we're done.
     *
     * (N.B. It's not clear that this helps that much.)
     */
    struct I {
        int pos;
        struct I* next;
    };

    struct hammer_arg* arg = (struct hammer_arg* )a;
    while (1) {
        pthread_mutex_lock(&glob);
        if (barrier_code >= T) {
            pthread_mutex_unlock(&glob);
            break;
        }
        pthread_mutex_unlock(&glob);
    }

    // Now fill an 8k block with our branches
    // at instr_block+(arg->proc & 0xffffffff)
    int n = 1 << 16;
    int i;
    struct I* locs[n];
    uint64_t code_ptr = (uint64_t) *(arg->proc);
    uint64_t hammer_ptr = (uint64_t) hammer_instr_block;
    uint64_t hammer_ptr_lower = hammer_ptr & 0xffffffffL;
    int64_t code_ptr_lower = (code_ptr & 0xffffffffL);
    int64_t k_ptr = MOD(code_ptr_lower - hammer_ptr_lower, BLOCK_SIZE);
    int64_t k = k_ptr / sizeof(int32_t);

    for (i = 0; i < n; i++) {
        struct I* s = (struct I*)malloc(sizeof(struct I));
        s->pos = 0;
        s->next = NULL;
        locs[i] = s;
    }

    int m;
    while (*arg->flag) {
        m = 1 + (rand() % (n - 1));

        for (i = 1; i < m; i++) {
            locs[i - 1]->next = locs[i];
        }

        locs[m - 1]->next = locs[0];

        shuffle(m, (void**)locs);

        for (i = 0; i < m; i++) {
            locs[i]->pos = i;
        }

        for (i = 0; i < m; i++) {
            int32_t offset = locs[i]->next->pos - i;
            uint32_t signed_offset = (uint32_t)offset & 0x3ffffff;
            
            hammer_instr_block[k+i] = (5 << 26) | signed_offset;
            
        }

        
        hammer_instr_block[k+m - 1] = 3596551104;  // RET
        

        delay();

        for (int i = 0; i < m; i++) {
        
            asm volatile (
                    "DC CVAU, %[jumper]\n"
                    "DSB ISH\n"
                    "IC IVAU, %[jumper]\n"
                    "DSB ISH\n"
            :
            : [jumper] "r" (hammer_instr_block+k+i)
            : "memory"
            );
        
        }

        asm volatile (
        
            "ISB\n"
            "BLR %[jumper]\n"
        
        :
        : [jumper] "r" (hammer_instr_block+k)
        :);

        delay();
    }


    return NULL;
}


int validate_results(struct Arg* arg) {
    int witnesses = 0;
    for (int n = 0; n < R; n++) {
        uint64_t p2x3 = *arg->results[n].p2x3;
        uint64_t p2x2 = *arg->results[n].p2x2;
        uint64_t p1x2 = *arg->results[n].p1x2;
        uint32_t lbl_a;
        asm volatile("adr x0, a\nldr %[data], [x0]\n" : [data] "=r" (lbl_a) : : "x0", "memory");
        
        uint32_t lbl_over;
        asm volatile("adr x0, over\nldr %[data], [x0]\n" : [data] "=r" (lbl_over) : : "x0", "memory");
        
        uint32_t lbl_end;
        asm volatile("adr x0, end\nldr %[data], [x0]\n" : [data] "=r" (lbl_end) : : "x0", "memory");
        
        uint32_t lbl_l1;
        asm volatile("adr x0, l1\nldr %[data], [x0]\n" : [data] "=r" (lbl_l1) : : "x0", "memory");
        
        uint32_t lbl_l2;
        asm volatile("adr x0, l2\nldr %[data], [x0]\n" : [data] "=r" (lbl_l2) : : "x0", "memory");
        
        uint32_t lbl_end2;
        asm volatile("adr x0, end2\nldr %[data], [x0]\n" : [data] "=r" (lbl_end2) : : "x0", "memory");
        
        
        if (p2x3 == 1) {
            
        if (p2x2 == 0) {
            
        if (p1x2 == 1) {
            witnesses++;
        }
        
        }
        
        }
        
    }
    return witnesses;
}

struct Arg* mkArg(void) {
    struct Arg* arg = (struct Arg*)malloc(sizeof(struct Arg) + sizeof(struct Result[R]));
    int** allcpus = (int**)malloc(sizeof(int*)*R);
    arg->cpus = allcpus;
int nprocs = get_nprocs();

    for (int k = 0; k < R; k++) {
        int* cpus = (int*)malloc(sizeof(int)*T);
        int start = rand() % nprocs;
        for (int i = 0; i < T; i++) {
            int top = rand() % 2;
            if (top == 0) {
                int b = start - (start % 8);
                start = b + rand() % 8;
            }
            else {
                int b = start - (start % 8);
                start = b + 8*(rand() % nprocs) + rand() % 8;
            }
            start = (nprocs + start) % nprocs;
            cpus[i] = start;
        }
        ishuffle(T, cpus);
        allcpus[k] = cpus;

        arg->results[k].p2x3 = (int64_t*) malloc(sizeof(int64_t));
        arg->results[k].p2x2 = (int64_t*) malloc(sizeof(int64_t));
        arg->results[k].p1x2 = (int64_t*) malloc(sizeof(int64_t));
        }
    shuffle(R, (void**)allcpus);

    int stride = rand() % R;
    arg->stride = stride;

    arg->mz = (uint64_t*)malloc(sizeof(uint64_t)*R);
    for (int i = 0; i < R; i++) {
        arg->mz[i] = 0;
    }
    

    return arg;
}

int main(int argc,char **argv) {
    // allocate a 512MiB space that we will fill with our branches at the designated locations
    hammer_instr_block = (uint32_t* )mmap(NULL, BLOCK_SIZE, PROT_READ|PROT_EXEC|PROT_WRITE, MAP_SHARED | MAP_ANONYMOUS, -1, 0);
    
    srand(time(NULL));

    if (argc > 1)
        R = atoi(argv[1]);
    else
        R = 1;

    int K = 3;
    struct Arg* arg = mkArg();

    barrier_init(T, &barrier);

    int flag = 1;
    int r;
    cpu_set_t hammer_cpus0;
    struct hammer_arg hammer_arg0;
    hammer_arg0.proc = &code_block_ptr0;
    hammer_arg0.flag = &flag;
    pthread_create(&hammer0, NULL, *hammer, (void* )&hammer_arg0);

    CPU_ZERO(&hammer_cpus0);
    CPU_SET(arg->cpus[0][0], &hammer_cpus0);
    if (r=pthread_setaffinity_np(hammer0, sizeof(cpu_set_t), &hammer_cpus0)) {
        errno=r;
        warn("setaffinity_hammer0");
    }
    cpu_set_t hammer_cpus1;
    struct hammer_arg hammer_arg1;
    hammer_arg1.proc = &code_block_ptr1;
    hammer_arg1.flag = &flag;
    pthread_create(&hammer1, NULL, *hammer, (void* )&hammer_arg1);

    CPU_ZERO(&hammer_cpus1);
    CPU_SET(arg->cpus[0][1], &hammer_cpus1);
    if (r=pthread_setaffinity_np(hammer1, sizeof(cpu_set_t), &hammer_cpus1)) {
        errno=r;
        warn("setaffinity_hammer1");
    }
    cpu_set_t hammer_cpus2;
    struct hammer_arg hammer_arg2;
    hammer_arg2.proc = &code_block_ptr2;
    hammer_arg2.flag = &flag;
    pthread_create(&hammer2, NULL, *hammer, (void* )&hammer_arg2);

    CPU_ZERO(&hammer_cpus2);
    CPU_SET(arg->cpus[0][2], &hammer_cpus2);
    if (r=pthread_setaffinity_np(hammer2, sizeof(cpu_set_t), &hammer_cpus2)) {
        errno=r;
        warn("setaffinity_hammer2");
    }
    
    

    DONE = 0;
    pthread_t timeout_t;
    pthread_create(&timeout_t, NULL, *timeout, NULL);

    void (*fps[T]);
    
    fps[0]=p0;
    fps[1]=p1;
    fps[2]=p2;
    

    shuffle(T, (void**)fps);
    shuffle(T, (void**)pts);
    for (int i = 0; i < T; i++) {
        pthread_create(pts[i], NULL, fps[i], (void* )arg);
    }

    shuffle(T, (void**)pts);
    for (int i = 0; i < T; i++) {
        pthread_join(*(pts[i]), NULL);
    }

    flag = 0;
    pthread_join(hammer0, NULL);
    pthread_join(hammer1, NULL);
    pthread_join(hammer2, NULL);
    
    
    DONE = 1;
    pthread_join(timeout_t, NULL);
    /* collect results */
    struct ResultPair count[R];
    int jfill = 0;
    int found = 0;
    for (int k = 0; k < R; k++) {
        for (int j = 0; j < jfill; j++) {
            if (*count[j].result->p2x3 != *arg->results[k].p2x3)
                 continue;
            if (*count[j].result->p2x2 != *arg->results[k].p2x2)
                 continue;
            if (*count[j].result->p1x2 != *arg->results[k].p1x2)
                 continue;
            count[j].count++;
            found = 1;
        }

        if (found == 0) {
            count[jfill].result = &(arg->results[k]);
            count[jfill].count=1;
            jfill++;
        }
        found = 0;
    }

    int witnesses = validate_results(arg);
    for (int k = 0; k < jfill; k++) {
        char* prefix = "";
        printf("{ %s \"p2x3\":%lu,\"p2x2\":%lu,\"p1x2\":%lu} : %lu\n", prefix,  *count[k].result->p2x3 ,  *count[k].result->p2x2 ,  *count[k].result->p1x2  , count[k].count);
    }
    printf("WITNESS: %d\n", witnesses);
    barrier_free(&barrier);
    munmap((void*)hammer_instr_block, BLOCK_SIZE);
    
    free(arg->mz);
    
    for (int k = 0; k < R; k++)
    {
        free(arg->results[k].p2x3);
        free(arg->results[k].p2x2);
        free(arg->results[k].p1x2);
        }
    free(arg);
    return 0;
}